@@ -56,6 +56,22 @@ impl Distribution<SplitHeuristic> for Standard {
     }
 }
 
+/// Heuristic for determining whether to prefer rotating cut pieces.
+#[derive(Copy, Clone, PartialEq, Eq)]
+pub(crate) enum RotateCutPieceHeuristic {
+    PreferUpright,
+    PreferRotated,
+}
+
+impl Distribution<RotateCutPieceHeuristic> for Standard {
+    fn sample<R: Rng + ?Sized>(&self, rng: &mut R) -> RotateCutPieceHeuristic {
+        match rng.gen_range(0..2) {
+            0 => RotateCutPieceHeuristic::PreferUpright,
+            _ => RotateCutPieceHeuristic::PreferRotated,
+        }
+    }
+}
+
 #[derive(Clone, Debug)]
 pub(crate) struct GuillotineBin {
     width: usize,
@@ -68,7 +84,11 @@ pub(crate) struct GuillotineBin {
 }
 
 impl Bin for GuillotineBin {
-    type Heuristic = (FreeRectChoiceHeuristic, SplitHeuristic);
+    type Heuristic = (
+        FreeRectChoiceHeuristic,
+        SplitHeuristic,
+        RotateCutPieceHeuristic,
+    );
 
     fn new(
         width: usize,
@@ -144,74 +164,182 @@ impl Bin for GuillotineBin {
             (
                 FreeRectChoiceHeuristic::BestAreaFit,
                 SplitHeuristic::ShorterLeftoverAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestAreaFit,
                 SplitHeuristic::LongerLeftoverAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestAreaFit,
                 SplitHeuristic::MinimizeArea,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestAreaFit,
                 SplitHeuristic::MaximizeArea,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestAreaFit,
                 SplitHeuristic::ShorterAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestAreaFit,
                 SplitHeuristic::LongerAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestShortSideFit,
                 SplitHeuristic::ShorterLeftoverAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestShortSideFit,
                 SplitHeuristic::LongerLeftoverAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestShortSideFit,
                 SplitHeuristic::MinimizeArea,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestShortSideFit,
                 SplitHeuristic::MaximizeArea,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestShortSideFit,
                 SplitHeuristic::ShorterAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestShortSideFit,
                 SplitHeuristic::LongerAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestLongSideFit,
                 SplitHeuristic::ShorterLeftoverAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestLongSideFit,
                 SplitHeuristic::LongerLeftoverAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestLongSideFit,
                 SplitHeuristic::MinimizeArea,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestLongSideFit,
                 SplitHeuristic::MaximizeArea,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestLongSideFit,
                 SplitHeuristic::ShorterAxis,
+                RotateCutPieceHeuristic::PreferUpright,
             ),
             (
                 FreeRectChoiceHeuristic::BestLongSideFit,
                 SplitHeuristic::LongerAxis,
+                RotateCutPieceHeuristic::PreferUpright,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestAreaFit,
+                SplitHeuristic::ShorterLeftoverAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestAreaFit,
+                SplitHeuristic::LongerLeftoverAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestAreaFit,
+                SplitHeuristic::MinimizeArea,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestAreaFit,
+                SplitHeuristic::MaximizeArea,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestAreaFit,
+                SplitHeuristic::ShorterAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestAreaFit,
+                SplitHeuristic::LongerAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestShortSideFit,
+                SplitHeuristic::ShorterLeftoverAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestShortSideFit,
+                SplitHeuristic::LongerLeftoverAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestShortSideFit,
+                SplitHeuristic::MinimizeArea,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestShortSideFit,
+                SplitHeuristic::MaximizeArea,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestShortSideFit,
+                SplitHeuristic::ShorterAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestShortSideFit,
+                SplitHeuristic::LongerAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestLongSideFit,
+                SplitHeuristic::ShorterLeftoverAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestLongSideFit,
+                SplitHeuristic::LongerLeftoverAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestLongSideFit,
+                SplitHeuristic::MinimizeArea,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestLongSideFit,
+                SplitHeuristic::MaximizeArea,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestLongSideFit,
+                SplitHeuristic::ShorterAxis,
+                RotateCutPieceHeuristic::PreferRotated,
+            ),
+            (
+                FreeRectChoiceHeuristic::BestLongSideFit,
+                SplitHeuristic::LongerAxis,
+                RotateCutPieceHeuristic::PreferRotated,
             ),
         ]
     }
@@ -221,7 +349,7 @@ impl Bin for GuillotineBin {
         cut_piece: &CutPieceWithId,
         heuristic: &Self::Heuristic,
     ) -> bool {
-        self.insert_with_heuristics(cut_piece, true, heuristic.0, heuristic.1)
+        self.insert_with_heuristics(cut_piece, true, heuristic.0, heuristic.1, heuristic.2)
     }
 
     fn insert_cut_piece_random_heuristic<R>(
@@ -234,6 +362,13 @@ impl Bin for GuillotineBin {
     {
         self.insert_cut_piece_with_heuristic(cut_piece, &rng.gen())
     }
+
+    fn matches_stock_piece(&self, stock_piece: &StockPiece) -> bool {
+        self.width == stock_piece.width
+            && self.length == stock_piece.length
+            && self.pattern_direction == stock_piece.pattern_direction
+            && self.price == stock_piece.price
+    }
 }
 
 impl GuillotineBin {
@@ -244,9 +379,12 @@ impl GuillotineBin {
         merge: bool,
         rect_choice: FreeRectChoiceHeuristic,
         split_method: SplitHeuristic,
+        rotate_preference: RotateCutPieceHeuristic,
     ) -> bool {
+        let prefer_rotated = rotate_preference == RotateCutPieceHeuristic::PreferRotated;
+
         if let Some((used_piece, free_index)) =
-            self.find_placement_for_cut_piece(cut_piece, rect_choice)
+            self.find_placement_for_cut_piece(cut_piece, rect_choice, prefer_rotated)
         {
             let free_rect = self.free_rects.swap_remove(free_index);
             self.split_free_rect_by_heuristic(&free_rect, &used_piece.rect, split_method);
@@ -267,6 +405,7 @@ impl GuillotineBin {
         &self,
         cut_piece: &CutPieceWithId,
         rect_choice: FreeRectChoiceHeuristic,
+        prefer_rotated: bool,
     ) -> Option<(UsedCutPiece, usize)> {
         let mut best_rect = Rect::default();
         let mut best_score = std::isize::MAX;
@@ -274,7 +413,7 @@ impl GuillotineBin {
         let mut free_index = None;
 
         for (i, free_rect) in self.free_rects.iter().enumerate() {
-            let fit = free_rect.fit_cut_piece(self.pattern_direction, cut_piece);
+            let fit = free_rect.fit_cut_piece(self.pattern_direction, cut_piece, prefer_rotated);
             match fit {
                 Fit::UprightExact => {
                     best_rect.x = free_rect.x;
@@ -464,14 +603,14 @@ impl GuillotineBin {
     }
 }
 
-impl Into<ResultStockPiece> for GuillotineBin {
-    fn into(self) -> ResultStockPiece {
-        ResultStockPiece {
-            width: self.width,
-            length: self.length,
-            pattern_direction: self.pattern_direction,
-            cut_pieces: self.cut_pieces.into_iter().map(Into::into).collect(),
-            waste_pieces: self.free_rects,
+impl From<GuillotineBin> for ResultStockPiece {
+    fn from(bin: GuillotineBin) -> Self {
+        Self {
+            width: bin.width,
+            length: bin.length,
+            pattern_direction: bin.pattern_direction,
+            cut_pieces: bin.cut_pieces.into_iter().map(Into::into).collect(),
+            waste_pieces: bin.free_rects,
         }
     }
 }
@@ -534,3 +673,151 @@ fn score_worst_short_side_fit(width: usize, length: usize, free_rect: &Rect) ->
 fn score_worst_long_side_fit(width: usize, length: usize, free_rect: &Rect) -> isize {
     -score_best_long_side_fit(width, length, free_rect)
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    #[test]
+    fn remove_cut_pieces() {
+        let cut_pieces = &[
+            CutPieceWithId {
+                id: 0,
+                external_id: None,
+                width: 10,
+                length: 10,
+                pattern_direction: PatternDirection::None,
+                can_rotate: false,
+            },
+            CutPieceWithId {
+                id: 1,
+                external_id: None,
+                width: 10,
+                length: 10,
+                pattern_direction: PatternDirection::None,
+                can_rotate: false,
+            },
+            CutPieceWithId {
+                id: 2,
+                external_id: None,
+                width: 10,
+                length: 10,
+                pattern_direction: PatternDirection::None,
+                can_rotate: false,
+            },
+            CutPieceWithId {
+                id: 3,
+                external_id: None,
+                width: 10,
+                length: 10,
+                pattern_direction: PatternDirection::None,
+                can_rotate: false,
+            },
+        ];
+
+        let heuristic = GuillotineBin::possible_heuristics()[0];
+
+        let mut bin = GuillotineBin::new(48, 96, 1, PatternDirection::None, 0);
+        cut_pieces.iter().for_each(|cut_piece| {
+            bin.insert_cut_piece_with_heuristic(cut_piece, &heuristic);
+        });
+
+        assert_eq!(bin.cut_pieces().len(), 4);
+
+        let cut_pieces_to_remove = [
+            UsedCutPiece {
+                id: 1,
+                external_id: None,
+                rect: Default::default(),
+                pattern_direction: PatternDirection::None,
+                is_rotated: false,
+                can_rotate: false,
+            },
+            UsedCutPiece {
+                id: 3,
+                external_id: None,
+                rect: Default::default(),
+                pattern_direction: PatternDirection::None,
+                is_rotated: false,
+                can_rotate: false,
+            },
+        ];
+
+        bin.remove_cut_pieces(cut_pieces_to_remove.iter());
+
+        assert_eq!(bin.cut_pieces().len(), 2);
+        assert_eq!(bin.cut_pieces().next().unwrap().id, 0);
+        assert_eq!(bin.cut_pieces().nth(1).unwrap().id, 2);
+    }
+
+    #[test]
+    fn bin_matches_stock_piece() {
+        let bin = GuillotineBin {
+            width: 48,
+            length: 96,
+            blade_width: 1,
+            pattern_direction: PatternDirection::None,
+            cut_pieces: Default::default(),
+            free_rects: Default::default(),
+            price: 0,
+        };
+
+        let stock_piece = StockPiece {
+            width: 48,
+            length: 96,
+            pattern_direction: PatternDirection::None,
+            price: 0,
+            quantity: Some(20),
+        };
+
+        assert!(bin.matches_stock_piece(&stock_piece));
+    }
+
+    #[test]
+    fn bin_does_not_match_stock_pieces() {
+        let bin = GuillotineBin {
+            width: 48,
+            length: 96,
+            blade_width: 1,
+            pattern_direction: PatternDirection::None,
+            cut_pieces: Default::default(),
+            free_rects: Default::default(),
+            price: 0,
+        };
+
+        let stock_pieces = &[
+            StockPiece {
+                width: 10,
+                length: 96,
+                pattern_direction: PatternDirection::None,
+                price: 0,
+                quantity: Some(20),
+            },
+            StockPiece {
+                width: 48,
+                length: 10,
+                pattern_direction: PatternDirection::None,
+                price: 0,
+                quantity: Some(20),
+            },
+            StockPiece {
+                width: 48,
+                length: 96,
+                pattern_direction: PatternDirection::ParallelToLength,
+                price: 0,
+                quantity: Some(20),
+            },
+            StockPiece {
+                width: 48,
+                length: 96,
+                pattern_direction: PatternDirection::None,
+                price: 10,
+                quantity: Some(20),
+            },
+        ];
+
+        stock_pieces
+            .iter()
+            .for_each(|stock_piece| assert!(!bin.matches_stock_piece(&stock_piece)))
+    }
+}
